import csv
import cv2
from pathlib import Path


import torch
from tqdm import tqdm
import numpy as np
import math

from navsim import NavSimGymEnv, DDPGAgent, Memory
from navsim.util import sizeof_fmt, image_layout
from navsim.util import ObjDict, ResourceCounter

import traceback
import numpy as np
from torch.utils.tensorboard import SummaryWriter


def s_hwc_to_chw(s):
    # TODO: HWC to CHW conversion optimized here
    # because pytorch can only deal with images in CHW format
    # we are making the optimization here to convert from HWC to CHW format.
    if isinstance(s, np.ndarray) and (s.ndim > 2):  # state is ndarray
        s = image_layout(s, 'hwc', 'chw')
    elif isinstance(s, list):  # state is a list of states
        for i in range(len(s)):
            if isinstance(s[i], np.ndarray) and (s[i].ndim > 2):  # state is ndarray
                s[i] = image_layout(s[i], 'hwc', 'chw')
    return s


class Executor:
    """
        TODO
    """

    def __init__(self, run_id='navsim_demo',
                 resume=True, conf=None):
        """

        :param conf: A ObjDict containing dictionary and object interface
        :param env: Any gym compatible environment
        :param resume: True: means continue if run exists, else start new
                           False: means overwrite if exists, else start new
        """
        self.run_id = run_id

        self.run_base_folder = Path(self.run_id)
        self.run_base_folder_str = str(self.run_base_folder.resolve())
        #        if run_base_folder.is_dir():
        #            raise ValueError(f"{run_base_folder_str} exists as a non-directory. "
        #                             f"Please remove the file or use a different run_id")
        if resume and self.run_base_folder.is_dir():
            self.conf = ObjDict().load_from_json_file(f"{self.run_base_folder_str}/conf.json")
            self.resume = True
            self.file_mode = 'a+'

        # else just start fresh
        else:
            self.conf = conf
            self.resume = False
            self.run_base_folder.mkdir(parents=True, exist_ok=True)
            self.conf.save_to_json_file(f"{self.run_base_folder_str}/conf.json")
            self.file_mode = 'w+'

        self.run_conf = ObjDict(self.conf.run_config)
        self.env_config = ObjDict(self.conf.env_config)

        pylog_filename = self.run_base_folder / 'py.log'  # TODO: use logger
        self.pylog_filename = str(pylog_filename.resolve())
        step_results_filename = self.run_base_folder / 'step_results.csv'
        self.step_results_filename = str(step_results_filename.resolve())
        episode_results_filename = self.run_base_folder / 'episode_results.csv'
        self.episode_results_filename = str(episode_results_filename.resolve())
        env_log_folder = self.run_base_folder / 'env.log'
        self.env_config.log_folder = str(env_log_folder.resolve())
        self.model_filename = f"{self.run_base_folder_str}/model_state.pt"
        self.memory_filename = f"{self.run_base_folder_str}/memory.pkl"
        # TODO: Add the code to delete previous files
        # TODO: Add the code to add categories
        self.summary_writer = SummaryWriter(f"{self.run_base_folder_str}/tb")

        self.rc = ResourceCounter()
        self.files_open()

        try:
            self.env = None
            self.env_open()
            if resume and self.run_base_folder.is_dir():
                self.memory = Memory.load_from_pkl(self.memory_filename)
            else:
                # TODO: HWC to CHW conversion optimized here
                # because pytorch can only deal with images in CHW format
                # we are making the optimization here to convert from HWC to CHW format.
                memory_observation_space_shapes = []
                for item in self.env.observation_space_shapes:
                    if len(item) == 3:  # means its an image in HWC
                        memory_observation_space_shapes.append((item[2], item[1], item[0]))
                    else:
                        memory_observation_space_shapes.append(item)
                self.memory = Memory(
                    capacity=self.run_conf.memory_capacity,
                    state_shapes=memory_observation_space_shapes,
                    action_shape=self.env.action_space.shape,
                    seed=self.run_conf['seed']
                )
                self.memory.info()

            self.max_action = self.env.action_space.high[0]
            self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

            self.agent = DDPGAgent(
                env=self.env,
                device=self.device,
                discount=self.run_conf['discount'], tau=self.run_conf['tau']
            )

            # ReplayBuffer(self.vector_state_dimension, self.action_dimension)

            self.apply_seed()

        # TODO: Enable tensorboard logging
        #        if self.enable_logging:
        #            from torch.utils.tensorboard import SummaryWriter
        #            self.writer = SummaryWriter('./logs/' + self.run_conf['env_name'] + '/')
        except Exception as e:
            self.env_close()
            self.files_close()
            print(traceback.format_exc())
            # print(e)

    # TODO: Find a better name for  this function
    def write_tb_values(self, values, t: int):
        for key, value in values.items():
            self.summary_writer.add_scalar(key, value, t)
            self.summary_writer.flush()

    def apply_seed(self):
        self.env.seed(self.run_conf['seed'])  # TODO: not needed because env is seeded at time of creation
        torch.manual_seed(self.run_conf['seed'])
        np.random.seed(self.run_conf['seed'])

    def files_open(self):
        self.pylog_file = open(self.pylog_filename, mode=self.file_mode)

        self.step_results_file = open(self.step_results_filename, mode=self.file_mode)
        self.step_results_writer = csv.writer(self.step_results_file,
                                              delimiter=',',
                                              quotechar='"',
                                              quoting=csv.QUOTE_MINIMAL)
        if not self.resume:
            self.step_results_writer.writerow(
                ['episode_num',
                 't',
                 'r',
                 'step_time',
                 'unity_step_time',
                 'peak_memory'])
        self.step_results_file.flush()

        self.episode_results_file = open(self.episode_results_filename, mode=self.file_mode)
        self.episode_results_writer = csv.writer(self.episode_results_file,
                                                 delimiter=',',
                                                 quotechar='"',
                                                 quoting=csv.QUOTE_MINIMAL)
        if not self.resume:
            self.episode_results_writer.writerow(
                ['episode_num',
                 'episode_reward',
                 'episode_time',
                 'episode_peak_memory'])
        self.episode_results_file.flush()

    def files_close(self):
        self.pylog_file.close()
        self.step_results_file.close()
        self.episode_results_file.close()

    def env_open(self):
        self.rc.start()
        self.env = NavSimGymEnv(self.env_config)
        self.env.reset()
        time_since_start, current_memory, peak_memory = self.rc.stop()
        log_str = f'Unity env creation resource usage: \n' \
                  f'time:{time_since_start},peak_memory:{sizeof_fmt(peak_memory)},' \
                  f'current_memory:{sizeof_fmt(current_memory)}\n'
        self.pylog_file.write(log_str)
        print(log_str)
        self.pylog_file.flush()
        self.env.info()
        #self.env.info_steps(save_visuals=True)

    def env_close(self):
        if self.env is None:
            print("Env is None")
        else:
            print("closing the env now")
            self.env.close()

    def execute(self):
        """
        Execute for the number of episodes
        :return:
        """
        t_max = int(self.run_conf['episode_max_steps'])
        num_episodes = int(self.run_conf['num_episodes'])
        checkpoint_interval = int(self.run_conf['checkpoint_interval'])
        num_episode_blocks = int(math.ceil(num_episodes / checkpoint_interval))

        for i in range(0,num_episode_blocks):
            for episode_num in tqdm(iterable=range((i*checkpoint_interval)+1, min((i+1)*checkpoint_interval,num_episodes)+1),
                                    desc=f"Episode {(i*checkpoint_interval)+1}-{min((i+1)*checkpoint_interval,num_episodes)}/{num_episodes}",
                                    unit='episode', ascii=True, ncols=80, bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} [{rate_fmt}{postfix}]'):
                self.rc.start()
                episode_resources = [self.rc.snapshot()]  # e0

                # initialise the episode counters
                episode_done = False
                episode_reward = 0
                t = 0

                self.env.start_navigable_map()
                # observe initial s
                s = self.env.reset()  # s comes out of env as a tuple always
                #self.env.step(self.env.action_space.sample())
                # get the navigable map and save it as image
                navigable_map = self.env.get_navigable_map()
                if navigable_map is not None:
                    cv2.imwrite(str((self.run_base_folder / f'navigable_map_{episode_num}.jpg').resolve()),navigable_map*255)
                else:
                    print(f'Map for episode {episode_num} is None')
                # TODO: HWC to CHW conversion optimized here
                # because pytorch can only deal with images in CHW format
                # we are making the optimization here to convert from HWC to CHW format.
                s = s_hwc_to_chw(s)

                samples_before_training = (self.run_conf['batch_size'] * self.run_conf['batches_before_train'])
                while not episode_done:
                    step_resources = [self.rc.snapshot()]  # s0
                    t += 1

                    # do the random sampling until enough memory is full
                    if self.memory.size < samples_before_training:
                        a = self.env.action_space.sample()
                    else:
                        # TODO: Find the best place to train, moved here for now
                        batch_s, batch_a, batch_r, batch_s_, batch_d = self.memory.sample(self.run_conf['batch_size'])
                        # print('training the agent')
                        self.agent.train(batch_s, batch_a, batch_r, batch_s_, batch_d)

                        a = (self.agent.select_action(s) + np.random.normal(
                            0, self.max_action * self.run_conf['expl_noise'],
                            size=self.env.action_space_shape[0])
                             ).clip(
                            -self.max_action,
                            self.max_action
                        )

                    step_resources.append(self.rc.snapshot())  # s1
                    s_, r, episode_done, info = self.env.step(a)

                    # TODO: HWC to CHW conversion optimized here
                    # because pytorch can only deal with images in CHW format
                    # we are making the optimization here to convert from HWC to CHW format.
                    s_ = s_hwc_to_chw(s_)

                    #    s = [[s]]  # make state a list of sequences
                    #    s_ = [[s_]]
                    # elif isinstance(s[0], (int,float)):  # state does not have multiple seq
                    # s = [s]
                    # s_ = [s_]
                    # for item in s_:

                    step_resources.append(self.rc.snapshot())  # s2

                    self.memory.append(
                        s=s, a=a, s_=s_, r=r,
                        d=float(episode_done))  # if t < t_max -1 else 1)
                    s = s_

                    episode_reward += r

                    # if self.memory.size >= self.run_conf['batch_size'] * self.run_conf['batches_before_train']:

                    #                    if (t >= self.config['batch_size'] * self.config['batches_before_train']) and (t % 1000 == 0):
                    # episode_evaluations.append(evaluate_policy(self.agent, self.env, self.config['seed']))

                    step_resources.append(self.rc.snapshot())  # 3

                    step_time = step_resources[3][0] - step_resources[0][0]
                    unity_step_time = step_resources[2][0] - step_resources[1][0]
                    peak_memory = step_resources[3][2]

                    # TODO: Collect these through tensorboard
                    self.step_results_writer.writerow(
                        [episode_num, t, r, step_time, unity_step_time, peak_memory])
                    self.step_results_file.flush()
                    if (t_max and t >= t_max):
                        break
                    # t += 1
                # end of while loop
                # episode end processing
                episode_resources.append(self.rc.stop())  # e1
                episode_time = episode_resources[1][0] - episode_resources[0][0]
                episode_peak_memory = episode_resources[1][2]
                # save every int(self.run_conf['checkpoint_interval'])
                # TODO: Save the episode_num
                if (episode_num % int(self.run_conf['checkpoint_interval'])) == 0:
                    self.agent.save_checkpoint(self.model_filename)
                    self.memory.save_to_pkl(self.memory_filename)

                #            if self.enable_logging:
                #                self.writer.add_scalar('Episode Reward', episode_reward, t)
                #            episode_rewards.append(episode_reward)
                self.write_tb_values({'reward': episode_reward,
                                      'time': episode_time,
                                      'memory': episode_peak_memory},
                                     episode_num)
                self.episode_results_writer.writerow([episode_num,
                                                      episode_reward,
                                                      episode_time,
                                                      episode_peak_memory])
                self.episode_results_file.flush()
                # s = self.env.reset()[0]
                # episode_done = False
                # episode_reward = 0
                # episode_timesteps = 0
                # episode_num += 1

        self.agent.save_to_onnx(folder=self.run_base_folder_str, critic=False)
        return



from typing import Union, Optional
import numpy as np
import pickle
import sys
from navsim.util import sizeof_fmt, image_layout


class Memory:
    """
    Stores and returns list of numpy arrays for s a r s_ d

    assumes you are storing s a r s_ d : state action reward next_state done
    # state / next_state : tuple of n-dim numpy arrays
    # a : 1-d numpy array of floats or ints
    # r : 1-d numpy array of floats #TODO: rewards could also be n-dim
    # d : 1-d numpy array of booleans

    # TODO: if we made this tensor based instead of np, would it make it better ?
    """

    @staticmethod
    def sample_info(s, a, r, s_, d):
        print("Sample Info")
        print('-----------')
        print("Shapes:")
        print('s:', [a.shape for a in s])
        print('a:', a.shape)
        print('r:', r.shape)
        print('s_:', [a.shape for a in s_])
        print('d:', d.shape)

    def __init__(self, capacity: int,
                 state_shapes: Union[list, tuple, int],
                 action_shape: Union[list, tuple, int],
                 seed: Optional[float] = None):
        """

        :param capacity: total capacity of the memory
        :param state_shapes: a tuple of state dimensions
        :param action_shape: int representing number of actions
        :param seed: seed to provide to numpy
        """
        self.capacity = capacity
        self.ptr: int = 0
        self.size: int = 0

        self.seed: Optional[float] = seed
        # from numpy.random import MT19937, RandomState, Generator
        # mt = MT19937(seed)
        # rs = RandomState(mt)
        # rg = Generator(mt)
        self.rng = np.random.default_rng(seed)

        # if isinstance(state_shapes, (list, tuple)):  # state_dims is an int
        if isinstance(state_shapes, (int, float)):  # state_dims is an int
            state_shapes = [[state_shapes]]  # make state_dims a list of sequences
        elif isinstance(state_shapes[0], (int, float)):  # state_dims first member is an int
            state_shapes = [state_shapes]  # make state_dims a list of sequences
        # else state_dims is a list of sequences, i.e. [[state_shape], [state_shap], [state_shape]]

        s_len = len(state_shapes)
        self.s = [None] * s_len
        self.s_ = [None] * s_len

        for i in range(s_len):
            state_shapes[i] = [capacity] + list(state_shapes[i])
            self.s[i] = np.full(state_shapes[i], 0.0)
            self.s_[i] = np.full(state_shapes[i], 0.0)

        action_shape = [capacity] + list(action_shape)
        self.a = np.full(action_shape, 0.0)

        self.r = np.full((capacity, 1), 0.0)
        self.d = np.full((capacity, 1), False)

    def append(self, s: Union[list, tuple, int, float],
               a: Union[list, tuple, int, float],
               r: float,
               s_: Union[list, tuple, int, float],
               d: int):
        if isinstance(s, (int, float)):  # state is an int
            s = [[s]]  # make state a list of sequences
            s_ = [[s_]]
        elif isinstance(s[0], (int, float)):  # state does not have multiple seq
            s = [s]
            s_ = [s_]
        for i in range(len(s)):
            self.s[i][self.ptr] = s[i]
            self.s_[i][self.ptr] = s_[i]

        self.a[self.ptr] = a
        self.r[self.ptr] = r
        self.d[self.ptr] = d

        self.ptr = (self.ptr + 1) % self.capacity
        self.size = min(self.size + 1, self.capacity)

    def sample(self, size):
        if (size == 0) or (size >= self.size):
            # idx = list(range(self.size))   # just return all
            return self.s, self.a, self.r, self.s_, self.d
        else:
            idx = self.rng.integers(low=0, high=self.size, size=size)
            s_len = len(self.s)
            return [self.s[i][idx] for i in range(s_len)], \
                   self.a[idx], \
                   self.r[idx], \
                   [self.s_[i][idx] for i in range(s_len)], \
                   self.d[idx]

    def info(self):
        print('-----------')
        print("Memory Info")
        print('-----------')
        print('capacity:', self.capacity)
        print('size:', self.size)
        print('seed:', self.seed)
        print("Shapes:")
        print('s:', [a.shape for a in self.s])
        print('a:', self.a.shape)
        print('r:', self.r.shape)
        print('s_:', [a.shape for a in self.s_])
        print('d:', self.d.shape)
        print("Sizes in bytes:")
        print('memory:', sizeof_fmt(sys.getsizeof(self)))
        print('s:', sizeof_fmt(sys.getsizeof(self.s)))
        print('a:', sizeof_fmt(sys.getsizeof(self.a)))
        print('r:', sizeof_fmt(sys.getsizeof(self.r)))
        print('s_:', sizeof_fmt(sys.getsizeof(self.s_)))
        print('d:', sizeof_fmt(sys.getsizeof(self.d)))

    def __len__(self):
        return len(self.mem)

    def __repr__(self):
        return str(self.mem)

    def __str__(self):
        return f'capacity={self.capacity} \nlength={len(self.mem)} \ncurrent pointer={self.ptr} \nmemory contents:\n{self.mem}'

    def __getitem__(self, sliced):
        return Memory(capacity=self.capacity, seed=self.seed, mem=self.mem[sliced])

    def save_to_pkl(self, filename):
        with open(filename, "wb") as out_file:
            pickle.dump(self, out_file)

    def save_to_npz(self, filename):
        pass

    @staticmethod
    def load_from_pkl(filename):
        with open(filename, "rb") as in_file:
            return pickle.load(in_file)

    def pretty_print(self, n_rows=0):
        """

        :param n_rows: 0 means all
        :return:
        """
        pass



import csv
from pathlib import Path
from typing import List, Any
import numpy as np
import uuid
import cv2

from typing import Any, Dict, List, Optional, Tuple, Union

from gym_unity.envs import (
    UnityToGymWrapper,
    GymStepResult
)

from mlagents_envs.environment import UnityEnvironment
from mlagents_envs.side_channel.side_channel import (
    SideChannel,
    IncomingMessage,
    OutgoingMessage,
)
from mlagents_envs.side_channel.engine_configuration_channel import \
    EngineConfigurationChannel
from mlagents_envs.side_channel.environment_parameters_channel import \
    EnvironmentParametersChannel

from mlagents_envs.side_channel.float_properties_channel import \
    FloatPropertiesChannel


def navsimgymenv_creator(env_config):
    return NavSimGymEnv(env_config)  # return an env instance


class NavSimGymEnv(UnityToGymWrapper):
    """NavSimGymEnv Class is a wrapper to Unity2Gym that inherits from the Gym interface

    The configuration provided is as follows:

    Observation Mode

    * Vector 0

    * Visual 1

    * VectorVisual 2

    Segmentation Mode

    * Object Segmentation 0

    * Tag Segmentation 1

    * Layer Segmentation 2

    Task

    * PointNav 0

    * SimpleObjectNav 1

    * ObjectNav 2

    Goal

    * 0 - Tocus

    * 1 - sedan1

    * 2 - Car1

    * 3 - Car2

    * 4 - City Bus

    * 5 - Sporty_Hatchback

    * Else - SEDAN

    .. code-block:: python

        env_config = ObjDict({
            "log_folder": "unity.log",
            "seed": 123,
            "timeout": 600,
            "worker_id": 0,
            "base_port": 5005,
            "observation_mode": 2,
            "segmentation_mode": 1,
            "task": 0,
            "goal": 0,
            "goal_distance":50
            "max_steps": 10,
            "reward_for_goal": 50,
            "reward_for_ep": 0.005,
            "reward_for_other": -0.1,
            "reward_for_falling_off_map": -50,
            "reward_for_step": -0.0001,
            "agent_car_physics": 0,
            "episode_max_steps": 10,
            "env_path":args["env_path"]
        })

    Action Space: [Throttle, Steering, Brake]

    * Throttle: -1.0 to 1.0

    * Steering: -1.0 to 1.0

    * Brake: 0.0 to 1.0

    Observation Space: [[Raw Agent Camera],[Depth Agent Camera],[Segmentation Agent Camera],[Agent Position, Agent Velocity, Agent Rotation, Goal Position]]
    The vector observation space:
        Agent_Position.x, Agent_Position.y, Agent_Position.z,
        Agent_Velocity.x, Agent_Velocity.y, Agent_Velocity.z,
        Agent_Rotation.x, Agent_Rotation.y, Agent_Rotation.z, Agent_Rotation.w,
        Goal_Position.x, Goal_Position.y, Goal_Position.z
    """

    def __init__(self, env_config) -> None:
        """
        env_config: The environment configuration dictionary or ObjDict Object
        """
        # filename: Optional[str] = None, observation_mode: int = 0, max_steps:int = 5):
        self.env_config = env_config
        self.observation_mode = self.env_config.get('observation_mode', 2)
        self.debug = env_config.get("debug", False)
        self.run_base_folder_str = env_config.get("run_base_folder_str", '.')
        self.run_base_folder = Path(self.run_base_folder_str)

        super().__init__(unity_env=self.__open_uenv(),
                         uint8_visual=False,
                         flatten_branched=False,
                         allow_multiple_obs=True)
        # (Env, uint8_visual, flatten_branched, allow_multiple_obs)
        # self.seed(self.conf['seed']) # unity-gym env seed is not working, seed has to be passed with unity env
        if self.debug:
            # open the debug files
            # TODO: Fix later
            self.file_mode = 'w+'
            self.actions_file = open(
                str((self.run_base_folder / 'actions.txt').resolve()),
                mode=self.file_mode)
            self.observations_file = open(
                str((self.run_base_folder / 'observations.txt').resolve()),
                mode=self.file_mode)
            self.actions_writer = csv.writer(self.actions_file,
                                             delimiter=',',
                                             quotechar='"',
                                             quoting=csv.QUOTE_MINIMAL)
            self.observations_writer = csv.writer(self.observations_file,
                                                  delimiter=',',
                                                  quotechar='"',
                                                  quoting=csv.QUOTE_MINIMAL)

    def __open_uenv(self) -> UnityEnvironment:
        # if self._env:
        #    raise ValueError('Environment already open')
        # else:
        log_folder = Path(self.env_config.get('log_folder', '.'))
        log_folder.mkdir(parents=True, exist_ok=True)

        engine_side_channel = EngineConfigurationChannel()
        environment_side_channel = EnvironmentParametersChannel()
        self.map_side_channel = MapSideChannel()
        self.float_properties_side_channel = FloatPropertiesChannel()

        # print(self.map_side_channel)
        engine_side_channel.set_configuration_parameters(time_scale=10,
                                                         quality_level=0)

        environment_side_channel.set_float_parameter(
            "rewardForGoalCollision",
            self.env_config.get('reward_for_goal', 50))
        environment_side_channel.set_float_parameter(
            "rewardForExplorationPointCollision",
            self.env_config.get('reward_for_ep', 0.005))
        environment_side_channel.set_float_parameter(
            "rewardForOtherCollision",
            self.env_config.get('reward_for_other', -0.1))
        environment_side_channel.set_float_parameter(
            "rewardForFallingOffMap",
            self.env_config.get('reward_for_falling_off_map', -50))
        environment_side_channel.set_float_parameter(
            "rewardForEachStep",
            self.env_config.get('reward_for_step', -0.0001))
        environment_side_channel.set_float_parameter(
            "segmentationMode", self.env_config.get('segmentation_mode', 1))
        environment_side_channel.set_float_parameter(
            "observationMode", self.env_config.get('observation_mode', 2))
        environment_side_channel.set_float_parameter(
            "episodeLength",
            self.env_config.get('episode_max_steps', 100))
        environment_side_channel.set_float_parameter(
            "selectedTaskIndex", self.env_config.get('task', 0))
        environment_side_channel.set_float_parameter(
            "goalSelectionIndex", self.env_config.get('goal', 0))
        environment_side_channel.set_float_parameter(
            "agentCarPhysics", self.env_config.get('agent_car_physics', 0))
        environment_side_channel.set_float_parameter(
            "goalDistance", self.env_config.get('goal_distance', 50))

        uenv_file_name = self.env_config.get('env_path')
        uenv_file_name = str(Path(uenv_file_name).resolve()) if \
            uenv_file_name is not None else None
        uenv = UnityEnvironment(file_name=uenv_file_name,
                                log_folder=str(log_folder.resolve()),
                                seed=self.env_config.get('seed'),
                                timeout_wait=self.env_config.get(
                                    'timeout', 600),
                                worker_id=self.env_config.get(
                                    'worker_id', 0),
                                # base_port=self.conf['base_port'],
                                no_graphics=False,
                                side_channels=[engine_side_channel,
                                               environment_side_channel,
                                               self.map_side_channel,
                                               self.float_properties_side_channel])

        return uenv

    def step(self, action: List[Any]) -> GymStepResult:
        result = super().step(action)

        if self.debug:
            self.actions_writer.writerow(action)
            self.actions_file.flush()
            if self.observation_mode in [0, 2]:
                vector_obs = result[0][-1]
                self.observations_writer.writerow(vector_obs)
                self.observations_file.flush()

        return result

    def info(self):
        """Prints the information about the environment

        """
        print('-----------')
        print("Env Info")
        print('-----------')
        if self.spec is not None:
            print(self.genv.spec.id)
        print('Action Space:', self.action_space)
        print('Action Space Shape:', self.action_space.shape)
        print('Action Space Low:', self.action_space.low)
        print('Action Space High:', self.action_space.high)
        print('Observation Mode:', self.observation_mode)
        # print('Gym Observation Space:', self.genv.observation_space)
        # print('Gym Observation Space Shape:', self.genv.observation_space.shape)
        print('Observation Space:', self.observation_space)
        print('Observation Space Shape:', self.observation_space.shape)
        print('Observation Space Shapes:', self.observation_space_shapes)
        print('Observation Space Types:', self.observation_space_types)
        print('Reward Range:', self.reward_range)
        print('Metadata:', self.metadata)
        return self

    def info_steps(self, save_visuals=False):
        """Prints the initial state, action sample, first step state

        """
        print('Initial State:', self.reset())
        action_sample = self.action_space.sample()
        print('Action sample:', action_sample)
        s, a, r, s_ = self.step(action_sample)
        print('First Step s,a,r,s_:', s, a, r, s_)
        if (self.observation_mode == 1) or (self.observation_mode == 2):
            for i in range(0, 3):
                cv2.imwrite(f'visual_{i}.jpg', (s[i] * 255).astype('uint8'))
        self.reset()
        return self

    @property
    def observation_space_shapes(self) -> list:
        """Returns the dimensions of the observation space
        """
        return [obs.shape for obs in self.observation_space.spaces]

    @property
    def observation_space_types(self) -> list:
        """Returns the dimensions of the observation space
        """
        return [type(obs) for obs in self.observation_space.spaces]

    def render(self, mode='') -> None:
        """Not implemented yet

        Args:
            mode:

        Returns:

        """
        pass

    def start_navigable_map(self, resolution_x=200, resolution_y=200,
                            cell_occupancy_threshold=0.5):
        """Get the Navigable Areas map

        Args:
            resolution_x: The size of the x axis of the resulting grid, default = 200
            resolution_y: The size of the y axis of the resulting grid, default = 200
            cell_occupancy_threshold: If at least this much % of the cell is occupied, then it will be marked as non-navigable, default = 50%

        Returns:
            A numpy array which has 0 for non-navigable and 1 for navigable cells

        Note:
            This method only works if you have called ``reset()`` or ``step()`` on the environment at least once.

        Note:
            Largest resolution that was found to be working was 2000 x 2000
        """
        self.map_side_channel.send_request("binaryMap",
                                           [resolution_x, resolution_y,
                                            cell_occupancy_threshold])
        # print('Inside get navigable map function:',self.map_side_channel.requested_map)

    def get_navigable_map(self) -> np.ndarray:
        """Get the Navigable Areas map

        Args:
            resolution_x: The size of the x axis of the resulting grid, default = 200
            resolution_y: The size of the y axis of the resulting grid, default = 200
            cell_occupancy_threshold: If at least this much % of the cell is occupied, then it will be marked as non-navigable, default = 50%

        Returns:
            A numpy array which has 0 for non-navigable and 1 for navigable cells

        Note:
            This method only works if you have called ``reset()`` or ``step()`` on the environment at least once.

        Note:
            Largest resolution that was found to be working was 2000 x 2000
        """
        return self.map_side_channel.requested_map

    def get_shortest_path_length(self):
        return self.float_properties_side_channel.get_property("ShortestPath")

    # Functions added to have parity with Env and RLEnv of habitat lab
    @property
    def sim(self):
        """Returns itself

        Added for compatibility with habitat API.

        Returns: link to self

        """
        return self

    @staticmethod
    def register_with_gym():
        """Registers the environment with gym registry with the name navsim

        """
        try:
            from gym.envs.registration import register
            register(id='navsim', entry_point='navsim.NavSimGymEnv')
        except Exception as e:
            print("Can not register NavSim Environment with Gym")
            print(e.message)

    @staticmethod
    def register_with_ray():
        """Registers the environment with ray registry with the name navsim

        """
        try:
            from ray.tune.registry import register_env
            register_env("navsim", navsimgymenv_creator)
        except Exception as e:
            print("Can not register NavSim Environment with Ray")
            print(e.message)


class MapSideChannel(SideChannel):
    """
    This is the SideChannel for retrieving map data from Unity.
    You can send map requests to Unity using send_request.
    The arguments for a mapRequest are ("binaryMap", [RESOLUTION_X, RESOLUTION_Y, THRESHOLD])
    """
    resolution = []

    def __init__(self) -> None:
        channel_id = uuid.UUID("24b099f1-b184-407c-af72-f3d439950bdb")
        super().__init__(channel_id)
        self.requested_map = None

    def on_message_received(self, msg: IncomingMessage) -> np.ndarray:
        if self.resolution is None:
            return None

        # mode as grayscale 'L' and convert to binary '1' because for some reason using only '1' doesn't work (possible bug)

        # img = Image.frombuffer('L', (self.resolution[0],self.resolution[1]), np.array(msg.get_raw_bytes())).convert('1')
        # timestr = time.strftime("%Y%m%d-%H%M%S")
        # img.save("img-"+timestr+".png")

        # raw_bytes = msg.get_raw_bytes()
        # unpacked_array = np.unpackbits(raw_bytes)[0:self.resolution[0]*self.resolution[1]]

        raw_bytes = msg.get_raw_bytes()
        self.requested_map = np.unpackbits(raw_bytes)[
                             0:self.resolution[0] * self.resolution[1]]
        self.requested_map = self.requested_map.reshape(
            (self.resolution[0], self.resolution[1]))
        # self.requested_map = np.array(msg.get_raw_bytes()).reshape((self.resolution[0],self.resolution[1]))
        # print('map inside on message received:',self.requested_map, self.requested_map.shape)
        return self.requested_map

        # np.savetxt("arrayfile", np.asarray(img), fmt='%1d', delimiter='')

    def send_request(self, key: str, value: List[float]) -> None:
        """
        Sends a request to Unity
        The arguments for a mapRequest are ("binaryMap", [RESOLUTION_X, RESOLUTION_Y, THRESHOLD])
        """
        self.resolution = value
        msg = OutgoingMessage()
        msg.write_string(key)
        msg.write_float32_list(value)
        super().queue_message_to_send(msg)



from pathlib import Path
from typing import Optional, List, Set

import attr
import cattr
import os
import navsim
from .util import ObjDict
import argparse
import yaml
from .cli_utils import argparser, non_default_args


@attr.s(auto_attribs=True)
class RunConfig:
    """
    Configuration for running in general
    """
    run_id: str = argparser.get_default("run_id")
    train = argparser.get_default("train")
    debug = argparser.get_default("num_envs")
    resume = argparser.get_default("resume")
    seed = argparser.get_default("force")


def main():
    """
    TODO: Implement configuration checks
    :return:
    """
    print(f'===================================')
    print(f'Navsim Version {navsim.__version__}')
    print(f'===================================')
    args = ObjDict(vars(argparser.parse_args()))
    print('arguments passed:')
    print(non_default_args)

    run_base_folder = Path(args.run_id)
    run_base_folder_str = str(run_base_folder.resolve())
    if args.resume and run_base_folder.is_dir():
        #TODO: Logic for resume here
        pass
    # else just start fresh
    else:
        run_base_folder.mkdir(parents=True, exist_ok=True)

    env_conf = ObjDict({
        "log_folder": str((run_base_folder / "unity.log").resolve()),
        "seed": int(args["seed"]),
        "timeout": 600,
        "worker_id": 0,
        "base_port": 5005,
        "observation_mode": int(args["observation_mode"]),
        "segmentation_mode": 1,
        "episode_max_steps": int(args["episode_max_steps"]) + 2,
        "task": 0,
        "goal": 0,
        "goal_distance": int(args["goal_distance"]),
        "reward_for_goal": 50,
        "reward_for_ep": 0.005,
        "reward_for_other": -0.1,
        "reward_for_falling_off_map": -50,
        "reward_for_step": -0.0001,
        "agent_car_physics": 0,
        "env_path": args["env_path"],
        "debug":args["debug"],
        "run_base_folder_str":run_base_folder_str
    })

    run_conf = ObjDict({
        "run_id": args["run_id"],
        "env_name": "navsim",
        "episode_max_steps": int(args["episode_max_steps"]),
        "num_episodes": int(args["num_episodes"]),
        "seed": int(args["seed"]),
        "discount": float(args["discount"]),
        "tau": float(args["tau"]),
        "expl_noise": float(args["exploration_noise"]),
        "memory_capacity": 100,
        "batch_size": int(args["batch_size"]),
        "batches_before_train": 2,
        "checkpoint_interval": int(args["checkpoint_interval"]),
    })

    if args["rl_backend"] == "rllib":
        import ray.rllib.agents.ppo as ppo
        conf = ppo.DEFAULT_CONFIG.copy()
        if args.debug:
            conf["log_level"] = "DEBUG"
        else:
            conf["log_level"] = "INFO"

        conf["framework"] = "torch"
        conf["ignore_worker_failures"] = True
    else:
        conf = ObjDict()
        conf['run_config'] = run_conf

    conf["env"] = "navsim"
    conf["num_workers"] = 1
    conf['env_config'] = env_conf


    print("Passed arguments + defaults:")
    print(args.to_yaml())
    #print("Final Configuration:")
    #print(conf.to_yaml())

    if args["rl_backend"] == "rllib":
        import ray
        ray.shutdown()
        ray.init(ignore_reinit_error=True)
        navsim.NavSimGymEnv.register_with_ray()
        result = ray.tune.run(
            ppo.PPOTrainer,
            config=conf,
            local_dir=run_base_folder_str,
            #stop={"episodes_total": run_conf.num_episodes}
            stop={"timesteps_total": 10},
            checkpoint_freq=1,
            checkpoint_at_end=True
        )
        best_checkpoint = result.get_last_checkpoint(
            metric="episode_reward_mean", mode="max"
        )
        #best_checkpoint = result.get_trial_checkpoints_paths(
        #    trial=result.get_best_trial("episode_reward_mean"),
        #    metric="episode_reward_mean", mode="max")

        print(best_checkpoint)
        trainer = ppo.PPOTrainer(config=conf)
        trainer.restore(best_checkpoint)
        model = trainer.get_policy().model
        print(type(model))
        print(model)
    else:
        executor = navsim.Executor(run_id=args["run_id"],
                                   resume=args["resume"],
                                   conf=conf)
        executor.execute()
        print("training finished")

        executor.env_close()
        print("env closed")
        executor.files_close()
        print("files closed")


# For python debugger to directly run this script
if __name__ == "__main__":
    main()
